#!/usr/bin/env python
##
##############################################################################
##
## Ubuntuzilla: install official Mozilla builds of Mozilla software on Ubuntu Linux
## Copyright (C) 2008  Daniel Folkinshteyn <nanotube@users.sf.net>
##
## http://ubuntuzilla.sourceforge.net/
##
## This program is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License
## as published by the Free Software Foundation; either version 3
## of the License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

##############################################################################
##
## Some notes about the general structure of this program.
##
## The idea is such that any action that is common to at least two of Firefox,
## Thunderbird, or Seamonkey is coded in the base class called 
## 'MozillaInstaller'. 
##
## Classes 'SeamonkeyInstaller', 'FirefoxInstaller', and 
## 'ThunderbirdInstaller' derive from the base 'MozillaInstaller' class, and 
## include all the package-specific actions. 
##
## This may seem a bit too complex, but it really simplifies code maintenance
## and the addition of new features, as it reduces the necessity of changing 
## the same code several times. 
##
## The 'BaseStarter' class processes the command line options, and decides
## what to do accordingly.
##
## The 'UbuntuzillaUpdater' class carries out updates of Ubuntuzilla itself.
##
## The 'VersionInfo' class is just a simple repository of version and other
## descriptive information about this software.
##
## The 'UtilityFunctions' class has some general functions that don't belong
## in the Mozilla classes.
##
## The 'Notifier' class is a wrapper for dbus Notifications that sends the
## GUI update notifications.
##
##############################################################################

from optparse import OptionParser
import optparse
import re
import os, os.path
import sys
import stat
import time
import shutil
import subprocess
import dbus
import urllib2
import traceback
import signal # used to workaround the python sigpipe bug

# todo: internationalization: figure out how to use the whole i18n thing, break out the text messages into separate files, and hopefully get some translators to work on those.

# some terminal escape sequences to make bold text
bold = "\033[1m"
unbold = "\033[0;0m"

class VersionInfo:
    '''Version information storage
    '''
    def __init__(self):
        self.name = "ubuntuzilla"
        self.version = "4.8.4"
        self.description = "Installer of Mozilla Builds of Mozilla Software for Ubuntu"
        self.url = "http://ubuntuzilla.sourceforge.net/"
        self.license = "GPL"
        self.author = "Daniel Folkinshteyn"
        self.author_email = "nanotube@users.sourceforge.net"
        self.platform = "Ubuntu Linux"

# Let's define some exceptions
class UbuntuzillaError(Exception): pass
class InsufficientDiskSpaceError(UbuntuzillaError): pass
class SystemCommandExecutionError(UbuntuzillaError): pass

class UtilityFunctions:
    '''This class is for holding some functions that are of general use, and thus
    do not belong in the mozilla class and its derivatives.
    '''
    
    def __init__(self, options):
        self.options=options
        self.version = VersionInfo()
    
    def getSystemOutput(self, executionstring, numlines=1, errormessage="Previous command has failed to complete successfully. Exiting."):
        '''Read output from an external command, exit if command fails.
        This is a simple wrapper for subprocess.Popen()
        For numlines==0, return whole list, otherwise, return requested number of lines.
        Result is a list, one line per item. 
        If numlines is 1, then result is a string.'''
        
        p = subprocess.Popen(executionstring, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, close_fds=True)
        returncode = p.wait()
        result = p.stdout.readlines()
        
        # need this separate check for w3m, since its return code is 0 even if it fails to find the site.
        if re.match("w3m", executionstring):
            if len(result) == 0 or re.match("w3m: Can't load", result[0]):
                errormessage = '\n'.join(result) + errormessage
                returncode = 1
        
        if returncode != 0:
            print >>sys.stderr, executionstring
            print >>sys.stderr, errormessage
            print >>sys.stderr, "Process returned code", returncode
            print >>sys.stderr, result
            raise SystemCommandExecutionError, "Command has not completed successfully. If this problem persists, please seek help at our website, " + self.version.url
        
        else:
            for i in range(0,len(result)):
                result[i] = result[i].strip()
            if numlines == 1:
                return result[0]
            elif numlines == 0:
                return result
            else:
                return result[0:numlines]
    
    def subprocess_setup(self):
        # Python installs a SIGPIPE handler by default. This is usually not what
        # non-Python subprocesses expect.
        # see http://bugs.python.org/issue1652
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)

    def execSystemCommand(self, executionstring, includewithtest=False, errormessage="Previous command has failed to complete successfully. Exiting."):
        '''Execute external command. Throw exception if command exits with non-zero status.
        This is a simple wrapper for subprocess.call()'''
        
        if (not self.options.test) or includewithtest:
            returncode = subprocess.call(executionstring, preexec_fn=self.subprocess_setup, shell=True)
            if returncode:
                print >>sys.stderr, executionstring
                print >>sys.stderr, errormessage
                print >>sys.stderr, "Process returned code", returncode
                raise SystemCommandExecutionError, "Command has not completed successfully. If this problem persists, please seek help at our website, " + self.version.url
    
    def robustDownload(self, argsdict, errormsg="Download failed. This may be due to transient network problems, so try again later. Exiting.", repeat=5, onexit = sys.exit):
        '''try the download several times, in case we get a bad mirror (happens 
        with a certain regularity), or some other transient network problem)
        
        note: repeat argument is not used anymore, we now iterate over mirror list'''
        
        #for i in xrange(repeat):
        origexecstring = argsdict['executionstring']
        for mirror in self.options.mirrors:
            try:
                argsdict['executionstring'] = re.sub("%mirror%",mirror,origexecstring)
                self.execSystemCommand(**argsdict)
                break
            except SystemCommandExecutionError:
                print "Error downloading. Trying again, hoping for a different mirror."
                time.sleep(2)
        else:
            print errormsg
            onexit(1)

class Notifier(object):
    '''A wrapper of the dbus Notifications service.
    
    We use it to send the gui update notifications.
    
    Code borrowed from:
    https://www.dfwpython.org/repo/Projects/DBUS/notifier.py
    
    See also here for dbus Notifications documentation:
    http://www.galago-project.org/specs/notification/0.9/x408.html
    '''

    SERVICE   = 'org.freedesktop.Notifications'
    OBJECT    = '/org/freedesktop/Notifications'
    INTERFACE = 'org.freedesktop.Notifications'

    def __init__(self):
        bus      = dbus.SessionBus()
        nfr_obj  = bus.get_object(Notifier.SERVICE, Notifier.OBJECT)
        self.nfr = dbus.Interface(nfr_obj, Notifier.INTERFACE)

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    def get_capabilities(self):
        return self.nfr.GetCapabilities()

    def notify(self, app_name='', replaces_id=0, app_icon='', summary='', 
                        body='', actionslist=[], hints={}, expire_timeout=0):
        '''Send a notification, with the specified parameters.
        
        app_name ::= The optional name of the application sending the
        notification. Can be blank.

        replaces_id ::= The optional notification ID that this notification
        replaces. The server must atomically (ie with no flicker or other
        visual cues) replace the given notification with this one. This allows
        clients to effectively modify the notification while it's active. A
        value of value of 0 means that this notification won't replace any
        existing notifications.

        app_icon ::= The optional program icon of the calling application. See
        Icons. Can be an empty string, indicating no icon.

        summary ::= The summary text briefly describing the notification.

        body ::= The optional detailed body text. Can be empty.

        actions ::= Actions are sent over as a list of pairs. Each even
        element in the list (starting at index 0) represents the identifier
        for the action. Each odd element in the list is the localized string
        that will be displayed to the user.

        hints ::= Optional hints that can be passed to the server from the
        client program. Although clients and servers should never assume each
        other supports any specific hints, they can be used to pass along
        information, such as the process PID or window ID, that the server may
        be able to make use of. See Hints. Can be empty.

        expire_timeout ::= The timeout time in milliseconds since the display
        of the notification at which the notification should automatically
        close.  If -1, the notification's expiration time is dependent on the
        notification server's settings, and may vary for the type of
        notification. If 0, never expire.
        '''
        return self.nfr.Notify(app_name, replaces_id, app_icon, summary, 
                                body, actionslist, hints, expire_timeout)


class BaseStarter:
    '''Parses options, and initiates the right actions.
    '''
    def __init__(self):
        self.version = VersionInfo()
        self.ParseOptions()
        
    def ParseOptions(self):
        '''Read command line options
        '''
        def prepend_callback(option, opt_str, value, parser):
            default_values = getattr(parser.values, option.dest)
            default_values.insert(0, value)
            setattr(parser.values, option.dest, default_values)
        
        parser = OptionParser(
                        version=self.version.name.capitalize() + " version " +self.version.version + "\nProject homepage: " + self.version.url, 
                        description="The Ubuntuzilla script can install the official Mozilla build of Firefox, Thunderbird, or Seamonkey, on an Ubuntu Linux system, in parallel with any existing versions from the repositories. For a more detailed usage manual, see the project homepage: " + self.version.url, 
                        formatter=optparse.TitledHelpFormatter(),
                        usage="%prog [options]\n or \n  python %prog [options]")
        parser.add_option("-d", "--debug", action="store_true", dest="debug", help="debug mode (print some extra debug output). [default: %default]")
        parser.add_option("-t", "--test", action="store_true", dest="test", help="make a dry run, without actually installing anything. [default: %default]")
        parser.add_option("-p", "--package", type="choice", action="store", dest="package", choices=['firefox','thunderbird','seamonkey'], help="which package to work on: firefox, thunderbird, or seamonkey. [default: %default]")
        parser.add_option("-a", "--action", type="choice", action="store", dest="action", choices=['install','remove','installupdater','removeupdater','checkforupdatetext','checkforupdategui','updateubuntuzilla'], help="what to do with the selected package: install, remove, installupdater, removeupdater, checkforupdatetext, checkforupdategui, or updateubuntuzilla. [default: %default]")
        parser.add_option("-g", "--skipgpg", action="store_true", dest="skipgpg", help="skip gpg signature verification. [default: %default]")
        parser.add_option("-u", "--unattended", action="store_true", dest="unattended", help="run in unattended mode. will not ask for user input. note that even in unattended mode, it will still ask for your input if you have a thunderbird profile both in ~/.thunderbird and ~/.mozilla-thunderbird (it's ok if the former is a symlink to the latter, though). [default: %default]")
        parser.add_option("-l", "--localization", action="store", dest="localization", help="for use with unattended mode only. choose localization (language) for your package of choice. note that the burden is on you to make sure that this localization of your package actually exists. [default: %default]")
        parser.add_option("-b", "--skipbackup", action="store_true", dest="skipbackup", help="skip the backup of the user profile for the package you are installing. if you already made a backup, or do not have sufficient disk space, use this option. [default: %default]")
        parser.add_option("-r", "--targetdir", action="store", dest="targetdir", help="installation/uninstallation target directory. [default: %default]")
        parser.add_option("-m", "--mirror", action="callback", callback=prepend_callback, type="string", dest="mirrors", help="Prepend a mozilla mirror server to the default list of mirrors. Use ftp mirrors only. Include path component up to the 'firefox', 'thunderbird', or 'seamonkey' directories. (See http://www.mozilla.org/mirrors.html for list of mirrors). [default: %default]")
        parser.add_option("-k", "--keyservers", action="callback", callback=prepend_callback, type="string", dest="keyservers", help="Prepend a pgp keyserver to the default list of keyservers. [default: %default]")
        
        parser.set_defaults(debug=False, 
                test=False, 
                package="firefox",
                action="install",
                skipgpg=False,
                unattended=False,
                localization="en-US",
                skipbackup=False,
                targetdir="/opt",
                mirrors=['mozilla.isc.org/pub/mozilla.org/',
                        'mozilla.ussg.indiana.edu/pub/mozilla.org/',
                        'ftp.osuosl.org/pub/mozilla.org/',
                        'mozilla.cs.utah.edu/pub/mozilla.org/',
                        'mozilla.mirrors.tds.net/pub/mozilla.org/',
                        'ftp.scarlet.be/pub/mozilla.org/',
                        'ftp.uni-erlangen.de/pub/mozilla.org/',
                        'sunsite.rediris.es/pub/mozilla.org/',
                        'www.gtlib.gatech.edu/pub/mozilla.org/',
                        'releases.mozilla.org/pub/mozilla.org/'],
                keyservers = ['subkeys.pgp.net',
                        'pgpkeys.mit.edu',
                        'pgp.mit.edu',
                        'wwwkeys.pgp.net',
                        'keymaster.veridis.com'])
        
        (self.options, args) = parser.parse_args()
        if self.options.debug:
            print "Your commandline options:\n", self.options
        
    def start(self):
        if self.options.action != 'updateubuntuzilla':
            self.check_uid()
            if self.options.package == 'firefox':
                fi = FirefoxInstaller(self.options)
                fi.start()
            elif self.options.package == 'thunderbird':
                ti = ThunderbirdInstaller(self.options)
                ti.start()
            elif self.options.package == 'seamonkey':
                si = SeamonkeyInstaller(self.options)
                si.start()
        else:
            ub = UbuntuzillaUpdater(self.options)
            ub.start()
    
    def check_uid(self):
        if os.getuid() == 0:
            print "\nYou appear to be trying to run Ubuntuzilla as root.\nUbuntuzilla really shouldn't be run as root under normal circumstances.\nYou are advised to exit now and run it as regular user, without 'sudo'.\nDo not continue, unless you know what you're doing.\nDo you want to exit now?"
            while 1:
                ans = raw_input("Please enter 'y' or 'n': ")
                if ans in ['y','Y','n','N']:
                    ans = ans.lower()
                    break
            
            if ans == 'y':
                print "Please run Ubuntuzilla again, without sudo."
                sys.exit()
            else:
                print "Hope you know what you're doing... Continuing..."
                    
    
class MozillaInstaller:
    '''Generic installer class, from which Firefox, Seamonkey, and Thunderbird installers will be derived.
    '''
    def __init__(self, options):
        self.options = options
        self.version = VersionInfo()
        self.util = UtilityFunctions(options)
        self.keySuccess = False
        if self.options.test:
            print "Testing mode ON."
        if self.options.debug:
            print "Debug mode ON."
        os.chdir('/tmp')
    
    def start(self):
        if self.options.action == 'install':
            self.install()
        elif self.options.action == 'remove':
            self.remove()
        elif self.options.action == 'installupdater':
            self.installupdater()
        elif self.options.action == 'removeupdater':
            self.removeupdater()
        elif self.options.action == 'checkforupdatetext':
            self.checkforupdateText()
        elif self.options.action == 'checkforupdategui':
            self.checkforupdateGui()
    
    def welcome(self):
        print "\nWelcome to the Ubuntuzilla " + self.options.package.capitalize() + " script version " + self.version.version + "\n\nUbuntuzilla is built only for Ubuntu Linux and other distributions derived from Ubuntu (such as Linux Mint, e.g.)\n\nThis script will now " + self.options.action + " the latest release of the official Mozilla build of " + self.options.package.capitalize() + " on your Linux system. If you run into any problems using this script, or have feature requests, suggestions, or general comments, please visit our website at", self.version.url, "\n"
        
    def setAptPackageName(self):
        '''Determine repositories package name for the sofware being installed.
        
        Since Ubuntuzilla likes to keep the repos versions installed, to make sure all the dependencies 
        are installed, and to let the user go back to the repos versions easily, if that is desired, 
        we need to know the repos package name.
        
        Due to the changes in package names over the various versions of Ubuntu, (e.g., mozilla-firefox 
        changed to firefox and mozilla-thunderbird changed to thunderbird), we need to "figure out" 
        the right package name.'''
        
        try:
            apt_cache_output = self.util.getSystemOutput(executionstring="apt-cache --names-only search '^" + self.options.package + "$' | grep '^" + self.options.package + "'")
            if apt_cache_output.startswith(self.options.package):
                self.aptPackage = self.options.package
        except (IndexError, SystemCommandExecutionError):
            try:
                apt_cache_output = self.util.getSystemOutput(executionstring="apt-cache --names-only search '^mozilla-"+self.options.package+"$' | grep '^mozilla-" + self.options.package + "'")
                if apt_cache_output.startswith('mozilla-'+self.options.package):
                    self.aptPackage = 'mozilla-'+self.options.package
            except (IndexError, SystemCommandExecutionError):
                while True:
                    self.aptPackage = raw_input('Unable to the repository package name for ' + self.options.package + '. Please look it up in your package manager and enter it manually here: ')
        
        if self.options.debug:
            print "Apt package name:", self.aptPackage, '\n'

    def getLatestVersion(self): # done in child, in self.releaseVersion
        print "Retrieving the version of the latest release of " + self.options.package.capitalize() + " from the Mozilla website..."
        # child-specific implementation comes in here

    def confirmLatestVersion(self):
        print bold + "The most recent release of " + self.options.package.capitalize() + " is detected to be " + self.releaseVersion + "." + unbold
        print "\nPlease make sure this is correct before proceeding. (You can confirm by going to http://www.mozilla.org/)"
        print "If no version number shows, if the version shown is not the latest, or if you would like to install a different release, press 'n', and you'll be given the option to enter the version manually. Otherwise, press 'y', and proceed with installation. [y/n]? "
        self.askyesno()
        if self.ans == 'y':
            pass
        else:
            print "\nIf no version shows, or it does not agree with the latest version as listed on http://www.mozilla.org, please visit our website at", self.version.url, "and let us know."
            print "If you would like to enter the version manually and proceed with installation, you can do so now. Note that beta and release candidate versions are now allowed, but you use pre-release software at your own risk!\n"
            
            while 1:
                self.ans = raw_input("Please enter the version of "+ self.options.package.capitalize() + " you wish to install, or 'q' to quit: ")
                if self.ans == 'q':
                    print 'Quitting by user request...'
                    sys.exit()
                else:
                    self.releaseVersion = self.ans
                    print "You have chosen version '" + self.releaseVersion + "'. Is that correct [y/n]?"
                    self.askyesno()
                    if self.ans == 'y':
                        break
                
    def getLocalizationList(self):
        if self.options.debug:
            tmp = os.popen("w3m -dump ftp://" + self.options.mirrors[0] + self.options.package + "/releases/" + self.releaseVersion + "/linux-i686/")
            print tmp.readlines()
            tmp.close()
        
        print "Retrieving list of available localizations for", self.options.package.capitalize(), "..."
        #for i in range(0, 5):
        for mirror in self.options.mirrors:
            try:
                self.locList = self.util.getSystemOutput(executionstring="w3m -dump ftp://" + mirror + self.options.package + "/releases/" + self.releaseVersion + "/linux-i686/ |grep '\. \. \.' |grep -v 'xpi'", numlines=0)
                break
            except SystemCommandExecutionError:
                print "Download error. Trying again, hoping for a different mirror."
                time.sleep(2)
        else:
            print "Failed to retrieve list of localizations. This may be due to transient network problems, so try again later. If the problem persists, please seek help on our website,", self.version.url
            sys.exit(1)
        
        # process the output to get the localization names
        # find the en-US line that we'll use to determine line format
        for line in self.locList:
            if re.search("en-US|en-GB|es-ES", line):
                break
        else:
            print "Some known localizations that are used to parse line format don't seem to exist. Please report this problem on our forums."
            print self.version.url
            sys.exit(1)
        
        # determine line format by looking at the en-US line.
        match = re.search('(.*)(en-US|en-GB|es-ES)(.*)', line)
        
        # process the lines
        self.locList = [re.sub(re.escape(match.group(1)) + r'([A-Za-z\-/]+) ?\. \. \. .*', r'\1', line.strip()).strip().strip('/') for line in self.locList]
    
    def get_localization_names(self):
        self.langsdict = {}
        try:
            response = urllib2.urlopen('http://www.mozilla.com/en-US/firefox/all.html')
            pagesource = response.read()
        except URLError:
            print "Unable to get localization names, proceeding without..."
            return
        
        match = re.search(r'var data = \[\[(.*?)\]\]', pagesource, re.M|re.S)
        try:
            langs = match.group(1)
            langs = langs.split('\n')
            langs = [lang.lstrip(' [').replace("'", "") for lang in langs]
        
            for lang in langs:
                lang = lang.split(', ')
                self.langsdict[lang[0]] = lang[1] + ' [' + lang[2] + ']'
        except KeyboardInterrupt:
            raise
        except:
            if self.options.debug:
                traceback.print_exc()
            print "Unable to get localization names, proceeding without..."
    
    def getLocalizationChoice(self):
        self.locChoice = len(self.locList)
        print "Please choose the localization (language) for " + self.options.package.capitalize() + ". Enter the number of your choice from the list below. [If you do not see a list of localizations, please check the help section of our website at", self.version.url, "for steps you can take to resolve this problem.]"
        for i in range(0,len(self.locList)):
            try:
                print "%3d. %-10s %-30s" % (i, self.locList[i], self.langsdict[self.locList[i]]), 
            except KeyError:
                print "%3d. %-10s %-30s" % (i, self.locList[i], ''),
            if (i+1)%1 == 0: # not really useful anymore
                print
        print
        
        if not self.options.unattended:
            while int(self.locChoice) not in range(0, len(self.locList)):
                # read choice, check if integer
                self.locChoice = raw_input("Enter your choice of localization (integer, from 0 to " + str(len(self.locList) - 1) + "): ")
                try:
                    self.locChoice = int(self.locChoice)
                except ValueError:
                    self.locChoice = len(self.locList)
        else:
            try:
                self.locChoice = self.locList.index(self.options.localization)
            except ValueError:
                print "Invalid localization supplied in the script command line argument. Please make sure to choose one of the available localizations from the list printed above."
                sys.exit(1)
        try:
            print "You have chosen localization " + bold + self.locList[self.locChoice] + ', ' + self.langsdict[self.locList[self.locChoice]] + unbold + ". Is that correct [y/n]? "
        except KeyError:
            print "You have chosen localization " + bold + self.locList[self.locChoice] + unbold + ". Is that correct [y/n]? "
        self.askyesno()
        if self.ans == 'y':
            return True
        else:
            print "In that case, you get another chance."
            return False

    def aptgetMeasures(self):
        
        print "\nUpdating repositories information with apt-get update. You may be prompted for your password.\n"
        returncode = os.system("sudo apt-get update")
        if returncode:
            print "It looks like repository updates have not finished successfully. This is most likely due to some repositories being temporarily down, or to a missing GPG key for a third-party repository (see detailed output above). If the main repositories look like they have been updated successfully, you may just proceed without worrying about this."
            print "Would you like to proceed with installation [y/n]? "
            self.askyesno()
            if self.ans == 'y':
                print "\nOK, Proceeding with installation."
            else:
                print "\nIn that case, see what's wrong with your repositories, and try again later."
                sys.exit(1)
        
        print "\nMaking sure that the repositories version of " + self.options.package.capitalize() + " is installed...\n"
        self.util.execSystemCommand(executionstring="sudo apt-get install " + self.aptPackage, errormessage="Package installation failed. Cannot proceed.")

    def checkDiskSpaceForBackup(self, target):
        requiredForBackup = int(self.util.getSystemOutput(executionstring="du -sk "+target+" |awk '{print $1}'"))
        availableForBackup = int(self.util.getSystemOutput(executionstring="df -k "+target+" | grep '[0-9]%' | awk '{ NN = NF - 2; print $NN }'"))
        if self.options.debug:
            print "availableForBackup, KB:", availableForBackup
            print "requiredForBackup, KB:", requiredForBackup
            print "Slack, KB:", availableForBackup - requiredForBackup
        if (availableForBackup - requiredForBackup > 1024*1024) or (float(requiredForBackup)/float(availableForBackup) < 0.10):
            return True
        else:
            raise InsufficientDiskSpaceError, "It has been determined that making a profile backup will leave less than 1 GB of free space, or take up more than 10% of the available free space. You should manually resolve this situation, and/or run with the '-b' option to skip backup. Please feel free to post a message on the Ubuntuzilla support forum if you have questions."

    def backupProfile(self): # done in child
        pass
    
    def downloadPackage(self): 
        # we are going to dynamically determine the package name
        print "Retrieving package name for", self.options.package.capitalize(), "..."
        for mirror in self.options.mirrors:
            try:
                self.packageFilename = self.util.getSystemOutput(executionstring="w3m -dump ftp://" + mirror + self.options.package + "/releases/" + self.releaseVersion + "/linux-i686/" + self.locList[self.locChoice] + "/ | grep '" + self.options.package + "' | grep -v '\.asc' |grep -v 'ftp://' | grep -v 'checksums' | awk '{ print substr($0,index($0, \"" + self.options.package + "\"))}' | awk '{print $1}' | sed -e 's/\.*$//'", numlines=1)
                print "Success!: " + self.packageFilename
                break
            except SystemCommandExecutionError:
                print "Download error. Trying again, hoping for a different mirror."
                time.sleep(2)
        else:
            print "Failed to retrieve package name. This may be due to transient network problems, so try again later. If the problem persists, please seek help on our website,", self.version.url
            sys.exit(1)
        
        
    def downloadGPGSignature(self): # done, self.sigFilename
        self.sigFilename = self.packageFilename + ".asc"
        print "\nDownloading " + self.options.package.capitalize() + " signature from the Mozilla site\n"
        
        self.util.robustDownload(argsdict={'executionstring':"wget -c --tries=5 --read-timeout=20 --waitretry=10 ftp://" + "%mirror%" + self.options.package + "/releases/" + self.releaseVersion + "/linux-i686/"+ self.locList[self.locChoice] + "/" + self.sigFilename, 'includewithtest':True}, errormsg="Failed to retrieve GPG key. This may be due to transient network problems, so try again later. Exiting.")
        
    def getMozillaGPGKey(self):
        ''' If key doesn't already exist on the system, retrieve key from keyserver.
        Try each keyserver in the list several times, sleep 2 secs between retries.'''
        
        # 812347DD - old mozilla software releases key
        # 0E3606D9 - current mozilla software releases key
        # 6CE2996F - mozilla messaging (thunderbird) key
        
        try:
            self.util.execSystemCommand("gpg --list-keys --with-colons 0E3606D9", includewithtest=True, errormessage="Mozilla GPG key not present on the system. Will attempt to retrieve from keyserver.")
            self.util.execSystemCommand("gpg --list-keys --with-colons 812347DD", includewithtest=True, errormessage="Mozilla GPG key not present on the system. Will attempt to retrieve from keyserver.")
            self.util.execSystemCommand("gpg --list-keys --with-colons 6CE2996F", includewithtest=True, errormessage="Mozilla GPG key not present on the system. Will attempt to retrieve from keyserver.")
        except SystemCommandExecutionError:
            print "\nImporting Mozilla Software Releases public key\n"
            print "Note that if you have never used gpg before on this system, and this is your first time running this script, there may be a delay of about a minute during the generation of a gpg keypair. This is normal and expected behavior.\n"
            
            for i in range(0,5):
                for keyserver in self.options.keyservers:
                    try:
                        self.util.execSystemCommand("gpg --keyserver " + keyserver + " --recv 0E3606D9 812347DD 6CE2996F", includewithtest=True)
                        self.keySuccess = True
                        print "Successfully retrieved Mozilla Software Releases Public key from", keyserver, ".\n"
                        break
                    except:
                        print "Unable to retrieve Mozilla Software Releases Public key from", keyserver, ". Trying again..."
                        time.sleep(2)
                if self.keySuccess:
                    break
            if not self.keySuccess:
                print "Failed to retrieve Mozilla Software Releases Public key from any of the listed keyservers. Please check your network connection, and try again later.\n"
                sys.exit(1)
    
    def getMD5Sum(self): # done, self.sigFilename
        self.sigFilename = self.packageFilename + ".md5"
        print "\nDownloading Seamonkey MD5 sums from the Mozilla site\n"
                
        self.util.robustDownload(argsdict={'executionstring':"wget -c --tries=5 --read-timeout=20 --waitretry=10 -q -nv -O - ftp://" + "%mirror%" + self.options.package + "/releases/" + self.releaseVersion + "/MD5SUMS | grep -F 'linux-i686/" + self.locList[self.locChoice] + "/" + self.packageFilename + "' > " + self.sigFilename, 'includewithtest':True}, errormsg="Failed to retrieve md5 sum. This may be due to transient network problems, so try again later. Exiting.")
        
        # example: 91360c07aea125dbc3e03e33de4db01a  ./linux-i686/en-US/seamonkey-2.0.tar.bz2
        # sed to:  91360c07aea125dbc3e03e33de4db01a  ./seamonkey-2.0.tar.bz2
        print "demunging: sed -i 's#linux-i686/" + self.locList[self.locChoice] + "/##' " + self.sigFilename + "...\n" 
        self.util.execSystemCommand("sed -i 's#linux-i686/" + self.locList[self.locChoice] + "/##' " + self.sigFilename, includewithtest=True)
        
    def verifyGPGSignature(self):
        print "\nVerifying signature...\nNote: do not worry about \"untrusted key\" warnings. That is normal behavior for newly imported keys.\n"
        returncode = os.system("gpg --verify " + self.sigFilename + " " + self.packageFilename)
        if returncode:
            print "Key verification failed. This is most likely due to a corrupt download. You should delete files '", self.sigFilename, "' and '", self.packageFilename, "' and run the script again.\n"
            print "Would you like to delete those two files now? [y/n]? "
            self.askyesno()
            if self.ans == 'y':
                print "\nOK, deleting files and exiting.\n"
                os.remove(self.packageFilename)
                os.remove(self.sigFilename)
            else:
                print "OK, exiting without deleting files.\n"
            sys.exit(1)
        
    def verifyMD5Sum(self):
        print "\nVerifying Seamonkey MD5 sum\n"
        returncode = os.system("md5sum -c " + self.sigFilename)
        if returncode:
            print "MD5 sum verification failed. This is most likely due to a corrupt download. You should delete files '", self.sigFilename, "' and '", self.packageFilename, "' and run the script again.\n"
            print "Would you like to delete those two files now? [y/n]? "
            self.askyesno()
            if self.ans == 'y':
                print "\nOK, deleting files and exiting.\n"
                os.remove(self.packageFilename)
                os.remove(self.sigFilename)
            else:
                print "OK, exiting without deleting files.\n"
            sys.exit(1)
        
    def extractArchive(self):
        print "\nExtracting archive\n"
        if re.search('\.tar\.gz$', self.packageFilename):
            self.tar_flags = '-xzf'
        elif re.search('\.tar\.bz2$', self.packageFilename):
            self.tar_flags = '-xjf'
        self.util.execSystemCommand(executionstring="sudo mkdir -p " + self.options.targetdir)
        if not self.options.test:
            self.util.execSystemCommand(executionstring="sudo tar -C " + self.options.targetdir + " " + self.tar_flags + " " + self.packageFilename)
        else:
            # in testing mode, extract to /tmp.
            self.util.execSystemCommand(executionstring="sudo tar -C " + '/tmp' + " " + self.tar_flags + " " + self.packageFilename, includewithtest=True)
        os.remove(self.packageFilename)
        if not self.options.skipgpg:
            os.remove(self.sigFilename)

    def linkPlugins(self):
        # order of preference:
        #/usr/lib/xulrunner-addons/plugins
        #/usr/lib/xulrunner-1.9a/plugins
        #/usr/lib/xulrunner/plugins
        #/usr/lib/firefox/plugins
        # releases after hardy don't need this at all...
        
        self.pluginPath = None
        
        print "Trying to determine firefox plugin path..."
                    
        result = self.util.getSystemOutput(executionstring="find /usr/lib -name 'libunixprintplugin.so'", numlines=0)
        for line in result:
            if re.match('/usr/lib/xulrunner-addons/plugins', line):
                self.pluginPath = os.path.dirname(line)
                break
        if self.pluginPath == None:
            for line in result:
                if re.search('/usr/lib/xulrunner\-[^/]/plugins', line):
                    self.pluginPath = os.path.dirname(line)
                    break
        if self.pluginPath == None:
            for line in result:
                if re.search('/usr/lib/xulrunner/plugins', line):
                    self.pluginPath = os.path.dirname(line)
                    break
        if self.pluginPath == None:
            for line in result:
                if re.search('/usr/lib/firefox/plugins', line):
                    self.pluginPath = os.path.dirname(line)
                    break
        
        if self.pluginPath == None:
            self.pluginPath = '/usr/lib/mozilla/plugins'
        
        print "Plugin path is: ", self.pluginPath
        
        print "\nLinking plugins\n"
        if os.path.lexists(os.path.join(self.options.targetdir, self.options.package, "plugins")):
            self.util.execSystemCommand(executionstring="sudo mv " + os.path.join(self.options.targetdir, self.options.package, "plugins") + " " + os.path.join(self.options.targetdir, self.options.package, "plugins_$(date -Iseconds)"))
        self.util.execSystemCommand(executionstring="sudo ln -s -f " + self.pluginPath + " " + os.path.join(self.options.targetdir, self.options.package, "plugins"))
        
        print "\nLinking dictionaries\n"
        if os.path.lexists(os.path.join(self.options.targetdir, self.options.package, "dictionaries")):
            self.util.execSystemCommand(executionstring="sudo mv " + os.path.join(self.options.targetdir, self.options.package, "dictionaries") + " " + os.path.join(self.options.targetdir, self.options.package, "dictionaries_$(date -Iseconds)"))
        self.util.execSystemCommand(executionstring="sudo ln -s -f " + "/usr/share/myspell/dicts" + " " + os.path.join(self.options.targetdir, self.options.package, "dictionaries"))
        
    def linkLauncher(self): # done in child
        pass
        
    def createMenuItem(self): # done in child
        pass
    
    def printSuccessMessage(self):
        print "\nThe new " + self.options.package.capitalize() + " version " + self.releaseVersion + " has been installed successfully."
        print bold + "\nMake sure to completely quit the old version of " + self.options.package.capitalize() + " for the change to take effect." + unbold

    def removeQuery(self):
        print "If the original Ubuntu repositories version of", self.options.package.capitalize(), "is a full version number older than the installed Mozilla build (e.g., 1.5 vs. 2.0), your profile may now be incompatible with the older", self.options.package.capitalize(), "version. If that happens, you can restore the backup profile that was made when you installed the Mozilla build."
        print "Are you sure you want to remove the official Mozilla build of", self.options.package.capitalize(), ", and revert to the Ubuntu repositories version [y/n]? "
        self.askyesno()
        if self.ans == 'y':
            if os.path.exists(os.path.normpath(self.options.targetdir + "/" + self.options.package)):
                pass
            else:
                print "\nIt appears that there's no Mozilla build installed in " + os.path.normpath(self.options.targetdir + "/" + self.options.package) + ". There is nothing to remove. Please make sure there is a Mozilla build installed on your system, and that you have specified the correct path."
                sys.exit(1)
        else:
            print "\nOK, exiting without taking any action."
            sys.exit()
        
    def removeLaunchLink(self): # done in child
        pass
        
    def removeBinaries(self):
        print "Would you like to completely delete the Mozilla build of", self.options.package.capitalize(), "from your hard drive [y/n]? "
        self.askyesno()
        if self.ans == 'y':
            self.util.execSystemCommand(executionstring="sudo rm -rf " + self.options.targetdir + "/" + self.options.package)
        else:
            print "\nOK, exiting without deleting the Mozilla build of", self.options.package.capitalize(), " install. If you wish to delete it manually later, it is located in " + self.options.targetdir + "/" + self.options.package + "."
    
    def printRemoveSuccessMessage(self):
        print "Successfully removed the Mozilla build of " + self.options.package.capitalize() + "."
    
    def getCurrentVersion(self): # done in child, self.installedVersion
        self.installedVersion = self.util.getSystemOutput(executionstring=self.options.package + """ --version | awk -F' ' '{for (field=1; field<=NF; ++field) {if (match($field, "^[0-9.]+,$")) {print $field}}}' | sed -e 's/,//'""", errormessage="Failed to determine current version. Exiting.")
    
    def updateActionsText(self): 
        if self.releaseVersion != self.installedVersion:
            print "The version of", self.options.package.capitalize(), "currently installed is", self.installedVersion
            print "The latest version available from Mozilla is", self.releaseVersion
            print "Please refer to the detailed instructions for updating", self.options.package.capitalize(), "on our site,", self.version.url + "#Update_Official_Mozilla_Build_of_" + self.options.package.capitalize(), "."
            print "\nIf you have not yet installed the official Mozilla build with the help of the Ubuntuzilla script, and are still running the repositories version, run this script with '-install' to install the latest Mozilla build of", self.options.package.capitalize(), "."
        else:
            print "The version of", self.options.package.capitalize(), "currently installed (" + self.installedVersion + ") matches the version available on the Mozilla site (" + self.releaseVersion + "). No update is necessary at this time."

    def install(self):
        self.welcome()
        self.setAptPackageName()
        self.getLatestVersion()
        self.confirmLatestVersion()
        self.getLocalizationList()
        self.get_localization_names()
        while not self.getLocalizationChoice():
            print "OK, try again"
        self.aptgetMeasures()
        self.backupProfile()
        self.downloadPackage()
        if not self.options.skipgpg:
            self.downloadGPGSignature()
            self.getMozillaGPGKey()
            self.verifyGPGSignature()
        self.getMD5Sum()
        self.verifyMD5Sum()
        self.extractArchive()
        self.linkPlugins()
        self.linkLauncher()
        self.createMenuItem()
        self.printSuccessMessage()
        self.installupdater()
        self.printSupportRequest()

    def remove(self):
        self.welcome()
        self.removeQuery()
        self.removeLaunchLink()
        self.removeBinaries()
        self.printRemoveSuccessMessage()
        self.removeupdater()
        self.printSupportRequest()
        
    def checkforupdateText(self):
        self.getLatestVersion()
        self.getCurrentVersion()
        self.updateActionsText()
    
    def checkforupdateGui(self):
        self.getLatestVersion()
        self.getCurrentVersion()
        self.updateActionsGui()
    
    def updateActionsGui(self):
        # print the versions for logging/troubleshooting purposes:
        print "Latest release version of", self.options.package.capitalize(), ":", self.releaseVersion
        print "Currently installed version of", self.options.package.capitalize(), ":", self.installedVersion
        
        if (self.releaseVersion != self.installedVersion) or self.options.debug:
            if self.options.package in ['firefox','thunderbird']:
                iconPath = self.options.targetdir + "/" + self.options.package + "/icons/mozicon50.xpm"
            else:
                iconPath = self.options.targetdir + "/" + self.options.package + "/chrome/icons/default/main-window.xpm"
            
            # grab the session bus, which is not available by default when
            # running from cron.
            homedir = os.environ['HOME']
            dbusfilelist = os.listdir(os.path.join(homedir, '.dbus/session-bus'))
            dbusfilelist = [fn for fn in dbusfilelist if os.path.isfile(os.path.join(homedir, '.dbus/session-bus', fn))]
            dbusfilelist.sort(key=lambda fn: os.path.getmtime(os.path.join(homedir, '.dbus/session-bus', fn)), reverse=True)
            if self.options.debug:
                print dbusfilelist
            try:
                dbusinfo = open(os.path.join(homedir, '.dbus/session-bus', dbusfilelist[0]),'r').read()
                if self.options.debug:
                    print dbusinfo
                sessionaddress = re.search(r'\nDBUS_SESSION_BUS_ADDRESS=([^\n]*)', dbusinfo).group(1)
                if self.options.debug:
                    print sessionaddress
                os.environ['DBUS_SESSION_BUS_ADDRESS'] = sessionaddress
            except KeyboardInterrupt:
                raise
            except:
                print >>sys.stderr, "ERROR: could not get DBUS_SESSION_BUS_ADDRESS, notification cannot be shown."
                return
            
            try:
                notifier = Notifier()
                notifier.notify(app_name = 'Ubuntuzilla', 
                                replaces_id = 0, 
                                app_icon = iconPath, 
                                summary = 'Ubuntuzilla: %s Update Available' % \
                                            self.options.package.capitalize(),
                                body = 'A '+self.options.package.capitalize()+' Update is Available. The version you are running is <b>'+self.installedVersion+'</b>. The new release available is <b>'+self.releaseVersion+'</b>. Please refer to the <a href="'+self.version.url+'#Update_Official_Mozilla_Build_of_' + self.options.package.capitalize() + '">detailed update instructions</a> on the Ubuntuzilla website.', 
                                expire_timeout=14400000)
            except KeyboardInterrupt:
                raise
            except: # could be just dbus.exceptions.DBusException?
                try:
                    self.util.execSystemCommand(executionstring="notify-send -i '" + iconPath + "' -t 14400000 'Ubuntuzilla: "+self.options.package.capitalize()+" Update Available.' 'A "+self.options.package.capitalize()+" Update is Available. The version you are running is <b>"+self.installedVersion+"</b>. The new release available is <b>"+self.releaseVersion+"</b>. Please refer to the <a href=\""+self.version.url+"#Update_Official_Mozilla_Build_of_" + self.options.package.capitalize() + "\">detailed update instructions</a> on the Ubuntuzilla website.'", errormessage="Failed to execute 'notify-send'. Please make sure that package 'libnotify-bin' is installed.")
                except KeyboardInterrupt:
                    raise
                except:
                    try:
                        import pygtk
                        pygtk.require('2.0')
                        import pynotify
                        if not pynotify.init('Ubuntuzilla'):
                            raise UbuntuzillaError
                        
                        n = pynotify.Notification('Ubuntuzilla: '+self.options.package.capitalize()+' Update Available.',
                                "A "+self.options.package.capitalize()+" Update is Available. The version you are running is <b>"+self.installedVersion+"</b>. The new release available is <b>"+self.releaseVersion+"</b>. Please refer to the <a href=\""+self.version.url+"#Update_Official_Mozilla_Build_of_" + self.options.package.capitalize() + "\">detailed update instructions</a> on the Ubuntuzilla website.",
                                'file://' + iconPath)
                        n.set_timeout(14400000)
                        
                        if not n.show():
                            raise UbuntuzillaError
                    except KeyboardInterrupt:
                        raise
                    except:
                        print >>sys.stderr, "ERROR showing GUI update notification."
            
        else:
            # exit silently, nothing needs to be done.
            sys.exit()
        
    def installupdater(self): 
        print "\nWould you like to set up automatic update checking for the official Mozilla build of "+self.options.package.capitalize()+" (recommended)? This feature will regularly check for updates to " + self.options.package.capitalize() + ", and put up a small unobtrusive notification message with update information, when a new release is available. [y/n] "
        self.askyesno()
        if self.ans == 'y':
            pass
        else:
            print "If you change your mind, you can always go back and enable the automatic update checker by running this script with option '-a installupdater'."
            sys.exit()
        
        #get rid of "no crontab for $username" message if empty crontab
        self.util.execSystemCommand(executionstring="crontab -l | crontab -")
        
        #get current crontab, grep out any old updater jobs
        #have to do regular os.system, since grep returns 1 when no matching lines are found, which is an expected occurrence.
        os.system("crontab -l | grep -v '^[^#].*ubuntuzilla.py -p "+self.options.package+"'> tempcrontabfile_autogenerated")
        
        #add cron job - run update checker every four (4) hours
        self.util.execSystemCommand(executionstring="echo '0 0,4,8,12,16,20 * * * /usr/local/bin/ubuntuzilla.py -p "+self.options.package+" -a checkforupdategui 2>&1 | logger -p user.notice -t \"UBUNTUZILLA\" -- ' >>tempcrontabfile_autogenerated")
        
        #set the crontab
        self.util.execSystemCommand(executionstring="crontab tempcrontabfile_autogenerated")
        
        #delete the temp file
        self.util.execSystemCommand(executionstring="rm -f tempcrontabfile_autogenerated")
        
        print "Updater successfully installed to your crontab. Try 'crontab -l' to see your new crontab, 'crontab -e' if you decide to edit it."
        #self.printSupportRequest()
        
    def printSupportRequest(self):
        print "\nThank you for using Ubuntuzilla.\n"
        print "Please support this project! If you found Ubuntuzilla helpful, please consider showing your appreciation with a small donation. :)\n"
        print "Just visit the project website,", self.version.url, ", and click the donate button."
    
    def removeupdater(self): 
        print "\nAre you sure you want to disable automatic update checking for the official Mozilla build of "+self.options.package.capitalize()+"? [y/n] "
        self.askyesno()
        if self.ans == 'y':
            print "If you change your mind, you can always go back and enable the automatic update checker by running this script with option '-a installupdater'."
        else:
            print "OK, leaving the update checking job as is."
            sys.exit()
        
        #get rid of "no crontab for $username" message if empty crontab
        self.util.execSystemCommand(executionstring="crontab -l | crontab -")
        
        #get current crontab, grep out any old updater jobs
        #have to do regular os.system, since grep returns 1 when no matching lines are found, which is an expected occurrence.
        os.system("crontab -l | grep -v '^[^#].*ubuntuzilla.py -p "+self.options.package+"'> tempcrontabfile_autogenerated")
        
        #set the crontab
        self.util.execSystemCommand(executionstring="crontab tempcrontabfile_autogenerated")
        
        #delete the temp file
        self.util.execSystemCommand(executionstring="rm -f tempcrontabfile_autogenerated")
        
        print "Updater successfully removed from your crontab. Try 'crontab -l' to see your new crontab, 'crontab -e' if you decide to edit it."
        #self.printSupportRequest()
        
    def askyesno(self):
        if not self.options.unattended:
            while 1:
                self.ans = raw_input("Please enter 'y' or 'n': ")
                if self.ans in ['y','Y','n','N']:
                    self.ans = self.ans.lower()
                    break
        else:
            self.ans = 'y'

class FirefoxInstaller(MozillaInstaller):
    '''This class works with the firefox package'
    '''
    def __init__(self,options):
        MozillaInstaller.__init__(self, options)

    def getLatestVersion(self):
        MozillaInstaller.getLatestVersion(self)
        self.releaseVersion = self.util.getSystemOutput(executionstring="wget -c --tries=20 --read-timeout=60 --waitretry=10 -q -nv -O - http://www.mozilla.com |grep 'product=' -m 1", numlines=1, errormessage="Failed to retrieve the latest version of "+ self.options.package.capitalize())
        self.releaseVersion = re.search(r'firefox\-(([0-9]+\.)+[0-9]+)',self.releaseVersion).group(1)
        

    def backupProfile(self):
        if os.path.lexists(os.path.expanduser('~/.mozilla')) and stat.S_ISDIR(os.lstat(os.path.expanduser('~/.mozilla'))[stat.ST_MODE]):
            if not self.options.skipbackup:
                print "\nBacking up old Firefox preferences\n"
                self.checkDiskSpaceForBackup("~/.mozilla")
                self.util.execSystemCommand(executionstring="cp -R ~/.mozilla ~/.mozilla_backup_$(date -Iseconds)", errormessage="Failed to back up profile.")
        else:
            print "\nOld firefox preferences not found. Nothing to back up. Proceeding with installation.\n"

    def downloadPackage(self): # done, self.packageFilename
        MozillaInstaller.downloadPackage(self)
        #self.packageFilename = self.options.package + "-" + self.releaseVersion + ".tar.gz"
        
        print "\nDownloading", self.options.package.capitalize(), "archive from the Mozilla site\n"
        
        self.util.robustDownload(argsdict={'executionstring':"wget -c --tries=5 --read-timeout=20 --waitretry=10 ftp://" + "%mirror%" + self.options.package + "/releases/" + self.releaseVersion + "/linux-i686/"+ self.locList[self.locChoice] + "/" + self.packageFilename, 'includewithtest':True})
    
    def getMD5Sum(self): #don't need, blank out
        pass
        
    def verifyMD5Sum(self): #don't need, blank out
        pass
    
    def linkLauncher(self):
        print "\nLinking launcher to new Firefox\n"
        self.util.execSystemCommand(executionstring="sudo dpkg-divert --divert /usr/bin/firefox.ubuntu --rename /usr/bin/firefox")
        self.util.execSystemCommand(executionstring="sudo ln -s -f "+self.options.targetdir+"/firefox/firefox /usr/bin/firefox")
        self.util.execSystemCommand(executionstring="sudo dpkg-divert --divert /usr/bin/mozilla-firefox.ubuntu --rename /usr/bin/mozilla-firefox")
        self.util.execSystemCommand(executionstring="sudo ln -s -f "+self.options.targetdir+"/firefox/firefox /usr/bin/mozilla-firefox")
    
    def removeLaunchLink(self):
        print "Removing /usr/bin links to the Mozilla build of Firefox...\n\n"
        self.util.execSystemCommand(executionstring="sudo rm -f /usr/bin/firefox")
        self.util.execSystemCommand(executionstring="sudo rm -f /usr/bin/mozilla-firefox")
        
        print "Rediverting /usr/bin links to point to the Ubuntu repositories version of Firefox...\n\n"
        self.util.execSystemCommand(executionstring="sudo dpkg-divert --rename --remove /usr/bin/mozilla-firefox")
        self.util.execSystemCommand(executionstring="sudo dpkg-divert --rename --remove /usr/bin/firefox")
        print "All your", self.options.package.capitalize(), "shortcuts now point to the Ubuntu repositories version of", self.options.package.capitalize(), "." 

    
class ThunderbirdInstaller(MozillaInstaller):
    '''This class works with the thunderbird package'
    '''
    def __init__(self,options):
        MozillaInstaller.__init__(self, options)

    def getLatestVersion(self):
        MozillaInstaller.getLatestVersion(self)
        self.releaseVersion = self.util.getSystemOutput(executionstring="wget -c --tries=20 --read-timeout=60 --waitretry=10 -q -nv -O - http://www.mozilla.com/thunderbird/ |grep 'product=' -m 1", numlines=1, errormessage="Failed to retrieve the latest version of "+ self.options.package.capitalize())
        self.releaseVersion = re.search(r'thunderbird\-(([0-9]+\.)+[0-9]+)',self.releaseVersion).group(1)

    def backupProfile(self):
        #if os.path.lexists(os.path.expanduser('~/.mozilla-thunderbird')) and stat.S_ISDIR(os.lstat(os.path.expanduser('~/.mozilla-thunderbird'))[stat.ST_MODE]):
        if os.path.isdir(os.path.expanduser('~/.mozilla-thunderbird')) and not os.path.islink(os.path.expanduser('~/.mozilla-thunderbird')):
            print "\nA Thunderbird profile has been found at ~/.mozilla-thunderbird. Do you want to make a backup of your Mozilla Thunderbird profile at ~/.mozilla-thunderbird? Note that if you store your email there, this may take quite a bit of disk space. You may wish to check the size of that directory and the amount of free disk space before you answer this question. Make backup? [y/n]: "
            self.askyesno()
            if self.ans == 'y':
                if not (self.options.unattended and self.options.skipbackup):
                    print "\nBacking up old thunderbird preferences\n"
                    self.checkDiskSpaceForBackup("~/.mozilla-thunderbird")
                    self.util.execSystemCommand(executionstring="cp -R ~/.mozilla-thunderbird ~/.mozilla-thunderbird-backup-$(date -Iseconds)")
            else:
                print "\nOK, proceeding without backup."

        #if os.path.lexists(os.path.expanduser('~/.thunderbird')) and stat.S_ISDIR(os.lstat(os.path.expanduser('~/.thunderbird'))[stat.ST_MODE]):
        if os.path.isdir(os.path.expanduser('~/.thunderbird')) and not os.path.islink(os.path.expanduser('~/.thunderbird')):
            print "\nA Thunderbird profile has been found at ~/.thunderbird. Do you want to make a backup of your Mozilla Thunderbird profile at ~/.thunderbird? Note that if you store your email there, this may take quite a bit of disk space. You may wish to check the size of that directory and the amount of free disk space before you answer this question. Make backup? [y/n]: "
            self.askyesno()
            if self.ans == 'y':
                if not (self.options.unattended and self.options.skipbackup):
                    print "\nBacking up old thunderbird preferences\n"
                    self.checkDiskSpaceForBackup("~/.thunderbird")
                    self.util.execSystemCommand(executionstring="cp -R ~/.thunderbird ~/.thunderbird-backup-$(date -Iseconds)")
            else:
                print "\nOK, proceeding without backup."

        #if os.path.lexists(os.path.expanduser('~/.mozilla-thunderbird')) and stat.S_ISDIR(os.lstat(os.path.expanduser('~/.mozilla-thunderbird'))[stat.ST_MODE]) and (not os.path.lexists(os.path.expanduser('~/.thunderbird')) or stat.S_ISLNK(os.lstat(os.path.expanduser('~/.thunderbird'))[stat.ST_MODE])):
        if os.path.isdir(os.path.expanduser('~/.mozilla-thunderbird')) and not os.path.islink(os.path.expanduser('~/.mozilla-thunderbird')) and ( not os.path.lexists(os.path.expanduser('~/.thunderbird')) or os.path.islink(os.path.expanduser('~/.thunderbird')) ):
            self.util.execSystemCommand(executionstring="ln -s -T -f ~/.mozilla-thunderbird ~/.thunderbird")

        #if os.path.lexists(os.path.expanduser('~/.thunderbird')) and stat.S_ISDIR(os.lstat(os.path.expanduser('~/.thunderbird'))[stat.ST_MODE]) and os.path.lexists(os.path.expanduser('~/.mozilla-thunderbird')) and stat.S_ISDIR(os.lstat(os.path.expanduser('~/.mozilla-thunderbird'))[stat.ST_MODE]):
        if os.path.isdir(os.path.expanduser('~/.thunderbird')) and not os.path.islink(os.path.expanduser('~/.thunderbird')) and os.path.isdir(os.path.expanduser('~/.mozilla-thunderbird')) and not os.path.islink(os.path.expanduser('~/.mozilla-thunderbird')):
            print "You seem to have two Thunderbird profiles, one in ~/.thunderbird and one in ~/.mozilla-thunderbird. Please choose among the following options. (If you need more information, please see our FAQ on this on our website,", self.version.url, "in the 'help on multiple profiles' section, prior to making this choice.)"
            print "1. use ~/.thunderbird directory as your profile for the new installation."
            print "2. use ~/.mozilla-thunderbird directory as your profile for the new installation (~/.thunderbird is moved out of the way, and a link to ~/.mozilla-thunderbird is created)"
            print "3. exit the script now, take your time and figure out which profile you want to use, and then try running this script again later when you are ready."
            self.ans=0
            while int(self.ans) not in range(1, 4):
                # read choice, check if numeric and integer
                self.ans = raw_input("Enter the number for your choice [1/2/3]: ")
                try:
                    self.ans = int(self.ans)
                except ValueError:
                    self.ans = 0
            
            if self.ans == 1:
                print "\nUsing ~/.thunderbird as your profile directory with no changes.\n"
            elif self.ans == 2:
                print "\nMoving your ~/.thunderbird directory to ~/.thunderbird-backup-$(date -Iseconds)\n"
                self.util.execSystemCommand(executionstring="mv ~/.thunderbird ~/.thunderbird-backup-$(date -Iseconds)")
                self.util.execSystemCommand(executionstring="ln -s -T -f ~/.mozilla-thunderbird ~/.thunderbird")
            elif self.ans == 3:
                print "Exiting. Run this script again when you figure out what profile to use."
                sys.exit(1)
        
        if not os.path.lexists(os.path.expanduser('~/.thunderbird')) and not os.path.lexists(os.path.expanduser('~/.mozilla-thunderbird')):
            self.util.execSystemCommand(executionstring="mkdir -p ~/.mozilla-thunderbird")
            self.util.execSystemCommand(executionstring="ln -s -T -f ~/.mozilla-thunderbird ~/.thunderbird")

    def downloadPackage(self): # done, self.packageFilename
        MozillaInstaller.downloadPackage(self)
        #self.packageFilename = self.options.package + "-" + self.releaseVersion + ".tar.gz"
        
        print "\nDownloading", self.options.package.capitalize(), "archive from the Mozilla site\n"
        
        self.util.robustDownload(argsdict={'executionstring':"wget -c --tries=5 --read-timeout=20 --waitretry=10 ftp://" + "%mirror%" + self.options.package + "/releases/" + self.releaseVersion + "/linux-i686/"+ self.locList[self.locChoice] + "/" + self.packageFilename, 'includewithtest':True})
    
    def getMD5Sum(self): #don't need, blank out
        pass
        
    def verifyMD5Sum(self): #don't need, blank out
        pass
        
    def linkPlugins(self):
        print "\nLinking dictionaries\n"
        if os.path.lexists(os.path.join(self.options.targetdir, self.options.package, "dictionaries")):
            self.util.execSystemCommand(executionstring="sudo mv " + os.path.join(self.options.targetdir, self.options.package, "dictionaries") + " " + os.path.join(self.options.targetdir, self.options.package, "dictionaries_$(date -Iseconds)"))
        self.util.execSystemCommand(executionstring="sudo ln -s -f " + "/usr/share/myspell/dicts" + " " + os.path.join(self.options.targetdir, self.options.package, "dictionaries"))
    
    def linkLauncher(self):
        print "\nLinking launcher to new thunderbird\n"
        self.util.execSystemCommand(executionstring="sudo dpkg-divert --divert /usr/bin/mozilla-thunderbird.ubuntu --rename /usr/bin/mozilla-thunderbird")
        self.util.execSystemCommand(executionstring="sudo dpkg-divert --divert /usr/bin/thunderbird.ubuntu --rename /usr/bin/thunderbird")
        self.util.execSystemCommand(executionstring="sudo ln -s -f "+self.options.targetdir+"/thunderbird/thunderbird /usr/bin/thunderbird")
        self.util.execSystemCommand(executionstring="sudo ln -s -f "+self.options.targetdir+"/thunderbird/thunderbird /usr/bin/mozilla-thunderbird")
        self.util.execSystemCommand(executionstring="sudo ln -s -f "+self.options.targetdir+"/thunderbird/thunderbird-bin "+self.options.targetdir+"/thunderbird/mozilla-thunderbird-bin")
    
    def removeLaunchLink(self):
        print "Removing /usr/bin links to the Mozilla build of Thunderbird...\n\n"
        self.util.execSystemCommand(executionstring="sudo rm -f /usr/bin/thunderbird")
        self.util.execSystemCommand(executionstring="sudo rm -f /usr/bin/mozilla-thunderbird")
        
        print "Rediverting /usr/bin links to point to the Ubuntu repositories version of Thunderbird...\n\n"
        self.util.execSystemCommand(executionstring="sudo dpkg-divert --rename --remove /usr/bin/mozilla-thunderbird")
        self.util.execSystemCommand(executionstring="sudo dpkg-divert --rename --remove /usr/bin/thunderbird")
        
        print "Resetting Preferred Applications setting back to default thunderbird\n\n"
        if os.path.lexists('/usr/bin/mozilla-thunderbird'):
            self.util.execSystemCommand(executionstring="gconftool-2 --set /desktop/gnome/url-handlers/mailto/command 'mozilla-thunderbird %s' --type 'string'")
        elif os.path.lexists('/usr/bin/thunderbird'):
            self.util.execSystemCommand(executionstring="gconftool-2 --set /desktop/gnome/url-handlers/mailto/command 'thunderbird %s' --type 'string'")

        print "All your", self.options.package.capitalize(), "shortcuts now point to the Ubuntu repositories version of", self.options.package.capitalize(), "." 
    

class SeamonkeyInstaller(MozillaInstaller):
    '''This class works with the seamonkey package'
    '''
    def __init__(self,options):
        MozillaInstaller.__init__(self, options)

    def aptgetMeasures(self): #don't need this for seamonkey (as of 4.2.0)
        pass

    def setAptPackageName(self):
        
        # override this for seamonkey, since we don't install 
        # seamonkey from the repos (it's not there in older versions of Ubuntu)
        # and thus don't need the repos package name.
        
        pass

    def getLatestVersion(self):
        MozillaInstaller.getLatestVersion(self)
        self.releaseVersion = self.util.getSystemOutput(executionstring="wget -c --tries=20 --read-timeout=60 --waitretry=10 -q -nv -O - http://www.seamonkey-project.org/ |grep 'product=' -m 1", numlines=1, errormessage="Failed to retrieve the latest version of "+ self.options.package.capitalize())
        self.releaseVersion = re.search(r'seamonkey\-(([0-9]+\.)+[0-9]+)',self.releaseVersion).group(1)
    
    def backupProfile(self):
        if os.path.lexists(os.path.expanduser('~/.mozilla')) and stat.S_ISDIR(os.lstat(os.path.expanduser('~/.mozilla'))[stat.ST_MODE]):
            if not self.options.skipbackup:
                print "\nBacking up old Seamonkey preferences\n"
                self.checkDiskSpaceForBackup("~/.mozilla")
                self.util.execSystemCommand(executionstring="cp -R ~/.mozilla ~/.mozilla_backup_$(date -Iseconds)")
        else:
            print "\nOld Seamonkey preferences not found. Nothing to back up. Proceeding with installation.\n"

    def downloadPackage(self): # done, self.packageFilename
        MozillaInstaller.downloadPackage(self)
        #self.packageFilename = self.options.package + "-" + self.releaseVersion + ".tar.gz"
        
        print "\nDownloading", self.options.package.capitalize(), "archive from the Mozilla site\n"
        
        self.util.robustDownload(argsdict={'executionstring':"wget -c --tries=5 --read-timeout=20 --waitretry=10 ftp://" + "%mirror%" + self.options.package + "/releases/" + self.releaseVersion + "/linux-i686/"+ self.locList[self.locChoice] + "/" + self.packageFilename, 'includewithtest':True})

    def downloadGPGSignature(self): #don't need this for seamonkey, blank it out
        pass

    def getMozillaGPGKey(self): #don't need this for seamonkey, blank it out
        pass
        
    def verifyGPGSignature(self): #don't need this for seamonkey, blank it out
        pass
        
    def linkPlugins(self):
        print "If you have the repositories version of Firefox installed, you can choose to link all the Firefox plugins to your Seamonkey install. Would you like to do that? [y/n]? "
        self.askyesno()
        if self.ans == 'y':
            MozillaInstaller.linkPlugins(self)
        
    def linkLauncher(self):
        
        print "\nCreating link to Seamonkey in /usr/bin/seamonkey\n"
        if os.path.exists('/usr/bin/seamonkey'):
            self.util.execSystemCommand(executionstring="sudo dpkg-divert --divert /usr/bin/seamonkey.ubuntu --rename /usr/bin/seamonkey")
        self.util.execSystemCommand(executionstring="sudo ln -s -f "+self.options.targetdir+"/seamonkey/seamonkey /usr/bin/seamonkey")

    def createMenuItem(self):
        print "Would you like to create an Applications menu item for Seamonkey? If you already have one, choose 'n'. [y/n]? "
        self.askyesno()
        if self.ans == 'y':
            print"Creating Applications menu item for Seamonkey.\n"
            menuitemfile = open("seamonkey.ubuntuzilla.desktop", "w+")
            menuitemfile.write('''[Desktop Entry]
Encoding=UTF-8
Name=Seamonkey
GenericName=Browser
Comment=Internet-Browser
Exec=seamonkey
Icon=''' + self.options.targetdir + '''/seamonkey/chrome/icons/default/seamonkey.png
Terminal=false
X-MultipleArgs=false
Type=Application
Categories=Application;Network;''')
            menuitemfile.close()
            self.util.execSystemCommand(executionstring="sudo mv -f seamonkey.ubuntuzilla.desktop /usr/share/applications")
            self.util.execSystemCommand(executionstring="sudo chown root:root /usr/share/applications/seamonkey.ubuntuzilla.desktop")
            self.util.execSystemCommand(executionstring="sudo chmod 644 /usr/share/applications/seamonkey.ubuntuzilla.desktop")
        else:
            print "OK, not creating menu item.\n"

    def printSuccessMessage(self):
        MozillaInstaller.printSuccessMessage(self)
        print "\nIf you are looking to use Seamonkey in multiple languages, head over to http://www.seamonkey-project.org/releases/#langpacks and download the installable language pack of your choice."
    
    def removeLaunchLink(self):
        print "Removing /usr/bin link to Seamonkey...\n\n"
        self.util.execSystemCommand(executionstring="sudo rm -f /usr/bin/seamonkey")
        if os.path.lexists('/usr/bin/seamonkey.ubuntu'):
            self.util.execSystemCommand(executionstring="sudo dpkg-divert --rename --remove /usr/bin/seamonkey")
        
        print "Removing Application menu item for Seamonkey...\n\n"
        self.util.execSystemCommand(executionstring="sudo rm -f /usr/share/applications/seamonkey.ubuntuzilla.desktop")

    
class UbuntuzillaUpdater:
    '''This class carries out the updates to Ubuntuzilla itself.
    We retrieve information about the latest version from a predefined text 
    file, and act accordingly.'''
    
    def __init__(self, options):
        self.version = VersionInfo()
        self.util = UtilityFunctions(options)
        self.options = options
        #self.updateInfoURL = "http://downloads.sourceforge.net/ubuntuzilla/ubuntuzillalatestreleaseinfo.txt"
        self.updateInfoURL = "http://ubuntuzilla.sourceforge.net/updates/ubuntuzillalatestreleaseinfo.txt"
        if self.options.test:
            self.updateInfoURL = "http://ubuntuzilla.sourceforge.net/ubuntuzillaupdatetest/ubuntuzillalatestreleaseinfo.txt"
        os.chdir('/tmp')
    
    def start(self):
        '''This is where we orchestrate the sequence of events...
        '''
        self.getUpdateInfo()
        if self.compareVersions():
            self.detectMachineHardware()
            self.installUpdate()
        
    def getUpdateInfo(self):
        '''Get the latest version information, and populate a dictionary with 
        it. The info text file is in a 'name=content' format, so that we don't
        have to worry about order of lines, and so that it is easily extensible
        to more items in the future (if needed).
        '''
        
        # try several times (in case we get a bad mirror, which happens with a certain regularity):
        for i in range(0, 5):
            try:
                self.updateinfo = self.util.getSystemOutput(executionstring="wget --tries=20 --read-timeout=60 --waitretry=10 -q -nv -O - " + self.updateInfoURL, numlines=0, errormessage="Failed to retrieve Ubuntuzilla update info. Detailed error information follows.")
                break
            except SystemCommandExecutionError:
                print "Error downloading latest release info. Trying again, hoping for a different mirror."
                time.sleep(2)
        else:
            print "Failed to retrieve Ubuntuzilla update info. This may be due to transient network problems, so try again later. Exiting."
            self.stop(1)
            
        if self.options.debug:
            print "Update information:", self.updateinfo
        
        self.updatedict = {}
        for item in self.updateinfo:
            name, content = item.split('=', 1)
            self.updatedict[name] = content
        
        if self.options.debug:
            print self.updatedict
        
    def compareVersions(self):
        # version numbers are of the form [0-9]+\.[0-9]+\.[0-9]+
        currentversionlist = [ int(i) for i in self.version.version.split('.') ]
        latestversionlist = [ int(i) for i in self.updatedict['releaseversion'].split('.') ]
        for i in range(0, len(currentversionlist)):
            if latestversionlist[i] > currentversionlist[i]:
                print "Ubuntuzilla update available..."
                return True
        else:
            print "Ubuntuzilla is already the latest version..."
            return False
    
    def installUpdate(self):
        '''get new ubuntuzilla_X.X.X.deb, check md5sum, install it with gdebi -n.
        unlike dpkg, gdebi will also take care of any dependencies, which is nice.
        '''
        
        # try several times (in case we get a bad mirror, which happens with a certain regularity):
        
        self.util.robustDownload(argsdict={'executionstring':"wget -c --tries=5 --read-timeout=20 --waitretry=10 -q -nv "+ self.updatedict['downloadurl'], 'includewithtest':True}, errormsg="Failed to download package. This may be due to transient network problems, so try again later. Exiting.", onexit = self.stop)
        
        self.util.robustDownload(argsdict={'executionstring':"wget -c --tries=5 --read-timeout=20 --waitretry=10 -q -nv "+ self.updatedict['md5sumurl'], 'includewithtest':True}, errormsg="Failed to download package MD5 sum. This may be due to transient network problems, so try again later. Exiting.", onexit=self.stop)
        
        try:
            self.util.execSystemCommand(executionstring="md5sum -c " + os.path.join("/tmp", os.path.split(self.updatedict['md5sumurl'])[1]), includewithtest=True)
        except SystemCommandExecutionError:
            print "Package MD5 sum failed verification. Deleting all downloaded files and exiting."
            self.stop(1)
        
        self.util.execSystemCommand(executionstring="sudo gdebi -n -q " + os.path.join("/tmp", os.path.split(self.updatedict['downloadurl'])[1]), includewithtest=True)
        print "Ubuntuzilla update installed successfully."
        self.stop(0)
        
    def detectMachineHardware(self):
        '''See if the OS is a 64bit or 32bit, so that we know to get the appropriate package.
        '''
        self.machineHardware = self.util.getSystemOutput(executionstring="uname -m")
        if self.options.debug:
            print "machineHardware:", self.machineHardware
        if self.machineHardware == 'x86_64' or self.machineHardware == 'ia64':
            self.machineHardware = 'amd64'
        else:
            self.machineHardware = 'i386'
            
        self.updatedict['downloadurl'] = re.sub(r"([0-9]+ubuntu[0-9]+)(\.deb)", r"\1" + "-" + self.machineHardware + r"\2", self.updatedict['downloadurl'])
        self.updatedict['md5sumurl'] = re.sub(r'([0-9]+ubuntu[0-9]+)(_md5sum\.txt)', r'\1' + "-" + self.machineHardware + r'\2', self.updatedict['md5sumurl'])
        
        if self.options.debug:
            print self.updatedict
        
    def stop(self, exitcode):
        '''Clean up after we are done and exit cleanly.
        '''
        try:
            os.remove(os.path.join("/tmp", os.path.split(self.updatedict['downloadurl'])[1]))
            os.remove(os.path.join("/tmp", os.path.split(self.updatedict['md5sumurl'])[1]))
        except:
            pass
        
        sys.exit(exitcode)

if __name__ == '__main__':
    
    bs = BaseStarter()
    bs.start()
